#!/usr/bin/env python

import unittest
import doctest
import sys
import os


def effect(msg):
    def callback(*args):
        print(msg)
    return callback


def doctest_AIController_choose_enemy():
    """Tests for AIController.choose_enemy.

        >>> from pyspacewar.ai import AIController
        >>> from pyspacewar.world import Ship, World, Object, Vector

    In a world with only two ships

        >>> ship = Ship()
        >>> other = Ship()
        >>> world = World()
        >>> world.add(ship)
        >>> world.add(other)
        >>> world.add(Object())

    the other ship is your enemy

        >>> ai = AIController(ship)
        >>> ai.choose_enemy() is other
        True

    If there are several enemy ships, let's engage the closest one

        >>> third = Ship()
        >>> world.add(third)
        >>> ship.position = Vector(10, 10)
        >>> other.position = Vector(-10, 5)
        >>> third.position = Vector(10, 15)
        >>> ai.choose_enemy() is third
        True

    but if we already have an enemy, keep after it

        >>> ai.enemy = other
        >>> ai.choose_enemy() is other
        True

    unles it goes too far away

        >>> other.position = Vector(-50, -50)
        >>> ai.choose_enemy() is third
        True

    If the closest enemy is already dead, switch to a live one

        >>> ai.enemy = third
        >>> third.dead = True
        >>> ai.choose_enemy() is other
        True

    """


def doctest_AIController_control():
    """Tests for AIController.control.

        >>> from pyspacewar.ai import AIController
        >>> from pyspacewar.world import Ship, World, Object, Vector
        >>> ship = Ship(velocity=Vector(30, 0))
        >>> world = World()
        >>> world.add(ship)
        >>> ai = AIController(ship)

    Not much to do in an empty world.

        >>> ai.control()
        >>> ship.engage_brakes
        True

    But if we find other ships, we can engage them!

        >>> other = Ship(position=Vector(100, 20))
        >>> world.add(other)
        >>> ai.control()

    unless we're dead

        >>> ship.dead = True
        >>> ai.control()

    """


def doctest_AIController_target():
    """Tests for AIController.target.

        >>> from pyspacewar.ai import AIController
        >>> from pyspacewar.world import Ship, World, Vector
        >>> ship = Ship(position=Vector(0, 0))
        >>> other = Ship(position=Vector(10, 20))
        >>> world = World()
        >>> world.add(ship)
        >>> world.add(other)
        >>> ai = AIController(ship)
        >>> ai.maybe_fire = effect('Foom!')

    The AI has to be able to turn the ship towards the enemy

        >>> ai.target(other)
        >>> ship.right_thrust
        0
        >>> ship.left_thrust > 0
        True

    We'll fire when we're turning across the enemy

        >>> ship.direction = 90
        >>> ai.target(other)
        Foom!

        >>> ship.direction = 45
        >>> ai.target(other)
        Foom!

    """


def doctest_AIController_evade():
    """Tests for AIController.evade.

        >>> from pyspacewar.ai import AIController
        >>> from pyspacewar.world import Ship, World, Missile, Planet, Vector
        >>> ship = Ship(position=Vector(0, 0))
        >>> other = Ship(position=Vector(10, 20))
        >>> sun = Planet(position=Vector(35, 20), radius=5)
        >>> world = World()
        >>> world.add(ship)
        >>> world.add(other)
        >>> world.add(sun)
        >>> ai = AIController(ship)

    In order not to crash into planets we need to take evasive action
    sometimes.  E.g. the planet is to the left of our path

        >>> ai.evade(other)
        >>> ship.left_thrust
        0
        >>> ship.right_thrust
        1

    and when it's to the right

        >>> ship.direction = 90
        >>> ai.evade(other)
        >>> ship.left_thrust
        1
        >>> ship.right_thrust
        0

    If we're further away from the planet, then we can follow the
    enemy instead of focusing 100% on the planet

        >>> ship.position = Vector(-100, -100)
        >>> ship.direction = 0
        >>> ship.left_thrust = 10
        >>> ship.right_thrust = 0
        >>> ai.evade(other)
        >>> ship.left_thrust
        10
        >>> ship.right_thrust
        1

        >>> ship.direction = 90
        >>> ship.left_thrust = 10
        >>> ship.right_thrust = 0
        >>> ai.evade(other)
        >>> ship.left_thrust
        11
        >>> ship.right_thrust
        0

    """


def doctest_AIController_get_closest_obstacle():
    """Tests for AIController.get_closest_obstacle.

        >>> from pyspacewar.ai import AIController
        >>> from pyspacewar.world import Ship, World, Missile, Planet, Vector
        >>> ship = Ship(position=Vector(30, 0))
        >>> missile = Missile(position=Vector(31, 0))
        >>> sun = Planet(position=Vector(35, 20), radius=5)
        >>> moon = Planet(position=Vector(0, -10), radius=2)
        >>> world = World()
        >>> world.add(ship)
        >>> world.add(missile)
        >>> world.add(sun)
        >>> world.add(moon)
        >>> ai = AIController(ship)

    In order not to crash into planets we need to pay attention

        >>> ai.get_closest_obstacle(ship) is sun
        True

    """


def doctest_AIController_maybe_fire():
    """Tests for AIController.maybe_fire.

        >>> from pyspacewar.ai import AIController
        >>> from pyspacewar.world import Ship, World, Object, Vector
        >>> ship = Ship()
        >>> ship.launch_effect = effect('Foom!')
        >>> world = World()
        >>> world.add(ship)
        >>> ai = AIController(ship)

    The AI always fires at very close range

        >>> enemy = Ship()
        >>> ai.maybe_fire(enemy, 10)
        Foom!

    unless the enemy is already dead

        >>> enemy.dead = True
        >>> ai.maybe_fire(enemy, 10)

    """


def test_suite():
    path = os.path.normpath(
        os.path.join(os.path.dirname(__file__), '..', '..'))
    if path not in sys.path:
        sys.path.append(path)
    return doctest.DocTestSuite()


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



"""
PySpaceWar version number tracker.
"""

version = "1.1.0"


def get_git_revision(unknown='', format=' (git %s)'):
    """Return the latest revision number of the files in the package."""
    import os
    import subprocess
    package_root = os.path.dirname(__file__)
    try:
        p = subprocess.Popen(['git', 'describe', '--always'],
                             cwd=package_root,
                             stdout=subprocess.PIPE)
    except OSError:  # pragma: nocover
        return unknown
    else:
        ver = p.communicate()[0].strip()
        if not isinstance(ver, str):
            ver = ver.decode('UTF-8', 'replace')
        return format % ver


if 'dev' in version:
    version += get_git_revision(format='+git.%s')



"""
Ship artificial intelligence.

Written by Ignas Mikalajunas.
"""

from .world import Ship


class AIController(object):
    """AI for a ship."""

    def __init__(self, ship):
        self.ship = ship
        self.rng = ship.world.rng
        self.last_l_r = 1
        self.enemy = None

    def choose_enemy(self):
        enemy = self.enemy
        if enemy is not None and enemy.dead:
            enemy = None
        if enemy is not None:
            dist_to_enemy = self.ship.distance_to(enemy)
        else:
            dist_to_enemy = 1e999
        threshold = 50
        for ship in self.ship.world.objects:
            if not isinstance(ship, Ship):
                continue
            if ship is self.ship or ship is self.enemy or ship.dead:
                continue  # pragma: nocover because peephole optimizer :/
            dist = self.ship.distance_to(ship)
            if enemy is None or dist < dist_to_enemy - threshold:
                enemy = ship
        return enemy

    def control(self):
        if self.ship.dead:
            return
        enemy = self.choose_enemy()
        self.enemy = enemy
        if enemy is not None:
            self.target(enemy)
        else:
            self.ship.brake()
        self.evade(enemy)

    def target(self, enemy):
        target_vector = enemy.position - self.ship.position
        moving_target_vector = target_vector + enemy.velocity - self.ship.velocity

        l_r = self.ship.direction_vector.cross_product(moving_target_vector)

        if target_vector.length() < 50:
            turn_const = 10
            thrust_const = 0
        else:
            turn_const = 2
            thrust_const = 1

        if l_r > 0:
            if self.last_l_r < 0:
                self.maybe_fire(enemy, target_vector.length())
            self.ship.left_thrust = self.rng.randrange(turn_const, turn_const + 5)
            self.ship.right_thrust = 0
        else:
            if self.last_l_r > 0:
                self.maybe_fire(enemy, target_vector.length())
            self.ship.left_thrust = 0
            self.ship.right_thrust = self.rng.randrange(turn_const, turn_const + 5)

        rel_velocity = (self.ship.velocity - enemy.velocity).length()
        if rel_velocity < 3:
            rel_velocity = 3
        if self.ship.velocity.length() ** 2 < self.rng.randrange(int(rel_velocity * 0.8), int(rel_velocity * 1.5) + 1) + 1:
            self.ship.forward_thrust = 1 * thrust_const
            self.ship.rear_thrust = 0
        else:
            self.ship.brake()

        self.last_l_r = l_r

    def evade(self, enemy):
        if enemy is not None and not enemy.dead:
            do_not_evade = enemy
        else:
            do_not_evade = None
        planet = self.get_closest_obstacle(self.ship, ignore=do_not_evade)
        if not planet:
            return

        evade_vector = planet.position - self.ship.position
        moving_target_vector = evade_vector - self.ship.velocity

        l_r = self.ship.direction_vector.cross_product(moving_target_vector)

        evade_vector_length = evade_vector.length()
        evade_factor = 1

        if evade_vector_length < 100:
            if l_r < 0:
                self.ship.left_thrust = evade_factor
                self.ship.right_thrust = 0
            else:
                self.ship.right_thrust = evade_factor
                self.ship.left_thrust = 0
        else:
            if l_r < 0:
                self.ship.left_thrust += evade_factor
            else:
                self.ship.right_thrust += evade_factor

    def get_closest_obstacle(self, what, ignore=None):
        distance = 300  # cutoff
        closest = None
        for obj in self.ship.world.objects:
            if obj.radius == 0:
                continue
            if obj is self.ship:
                continue
            if obj is ignore:
                continue
            dst = what.distance_to(obj)
            if dst < distance:
                closest = obj
                distance = dst
        return closest

    def maybe_fire(self, enemy, distance):
        if not enemy.dead:
            if 0 == self.rng.randrange(0, int(distance / 30) + 1):
                self.ship.launch()



#!/usr/bin/env python

import unittest
import doctest
import os
import sys


class FakeGameUI(object):

    counter = 3

    def load_settings(self):
        pass

    def init(self):
        pass

    def wait_for_tick(self):
        pass

    def interact(self):
        self.counter -= 1
        if self.counter <= 0:
            sys.exit()

    def draw(self):
        pass

    def save_settings(self):
        pass


def doctest_main():
    """Test for main

        >>> from pyspacewar.main import main
        >>> main(['-f', '-d', '--no-sound', '--no-music', '-m', '640x480'])

        >>> main(['-m', 'lalala'])
        Traceback (most recent call last):
          ...
        SystemExit: pyspacewar: error: invalid mode: lalala

    """


def setUp(test=None):
    from pyspacewar import main
    main.GameUI = FakeGameUI


def test_suite():
    path = os.path.normpath(
        os.path.join(os.path.dirname(__file__), '..', '..'))
    if path not in sys.path:
        sys.path.append(path)
    return doctest.DocTestSuite(setUp=setUp)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#!/usr/bin/env python
import os
import sys

from setuptools import setup


with open('NEWS.rst') as f:
    news = f.read()

long_description = """\
Two ships duel in a gravity field.   Gravity doesn't affect
the ships themselves (which have spanking new anti-gravity
devices), but it affects missiles launced by the ships.

You can play against the computer, or two players can play
with one keyboard.  There is also a Gravity Wars mode, where
the two ships do not move, and the players repeatedly
specify the direction and velocity of their missiles.

Latest changes
--------------

""" + '\n\n'.join(news.split('\n\n')[:2])

pkgdir = os.path.join('src', 'pyspacewar')


def determine_version():
    sys.path.insert(0, 'src')
    from pyspacewar.version import version
    del sys.path[0]
    return version
version = determine_version()

setup(
    name='pyspacewar',
    version=version,
    author='Marius Gedminas',
    author_email='marius@gedmin.as',
    license='GPL',
    platforms=['any'],
    url='http://mg.pov.lt/pyspacewar/',
    description='A game loosely inspired by the original Spacewar',
    long_description=long_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Games/Entertainment :: Arcade',
    ],
    scripts=['pyspacewar'],
    packages=['pyspacewar'],
    package_dir={'': 'src'},
    package_data={
        'pyspacewar': [
            'icons/*',
            'images/*',
            'music/*',
            'sounds/*',
        ],
    },
    install_requires=['pygame'],
)
